from gearup import *

def test_kwargs():
  from gearup import gearup

  def f(**kwargs):
    print(kwargs)

  print()
  gearup(f)(['clf.alpha=1', 'clf.beta=2', 'method.beta=3'])

  def f1(alpha: float): return alpha
  def f2(beta: float, gamma: float): return beta + gamma

  def main(f, **kwargs):
    if f == 'f1':
      return apply(f1)(**kwargs.get('func', dict()))
    elif f == 'f2':
      return apply(f2)(**kwargs.get('func', dict()))
    else:
      raise NotImplementedError()

  gearup(main)(['f=f1', 'func.alpha=3'])
  gearup(main)(['f=f2', 'func.beta=5', 'func.gamma=6'])

def test_conf():
  def f1(alpha: float): return 2 * alpha
  def f2(beta: float, gamma: float): return beta + gamma

  def g1(x: float): return x + 1
  def g2(x: float, y: float): return x + y

  def main(f, g, **kwargs):
    return apply(f)(**kwargs['fargs']) * apply(g)(**kwargs['gargs'])

  assert gearup(main)(['f=f1', 'g=g2', 'fargs.alpha=2', 'gargs.x=2.0', 'gargs.y=1.5']) == 14.0
  assert gearup(main)(['f=f2', 'g=g1', 'fargs.beta=2', 'fargs.gamma=1e-1', 'gargs.x=9.0']) == 21.0

  try:
    gearup(main)(['f=f1'])
  except (ValueError, TypeError):
    import traceback
    print(traceback.format_exc())
  else:
    raise Exception()


__all__ = [
  'indent',
  'indent_second'
]

def indent(string, n=2):
  tokens = string.split('\n')
  ind = ' ' * n

  return '\n'.join(
    '%s%s' % (ind, token)
    for token in tokens
  )

def indent_second(string, n=2):
  tokens = string.split('\n')
  sep = '\n%s' % (' ' * n, )

  return sep.join(tokens)


from gearup import apply

def test_apply():
  class A(object):
    def __init__(self, x):
      self.x = x

  def f(a: A, x: int, y: int):
    return a

  assert isinstance(apply(f, '1', 2, 3), A)
  assert isinstance(apply(f, '1', 2, 3).x, str)
  assert isinstance(apply(f, A(1), 2, 3).x, int)


from gearup import *

def test_gear_up():
  def f(x : int, y : int):
    return x + y

  def g1(w : int, y : int):
    return w + y

  def g2(t : int, y : int):
    return t + y

  assert gearup(
    train=f,
    test=dict(
      x1=g1,
      x2=g2
    )
  )(['test', 'x1', 'w=2', 'y=4']) == 6

def test_choice():
  def f(x: int):
    return x

  def f2(x: int):
    return x

  def f3(x: int):
    return x + 1

  def main(method: choice(f, g=f2, h=f3), x: int):
    return method(x)

  gearup(main)(['f', '1'])
  gearup(main)(['g', '1'])
  gearup(main)(['h', '1'])

  try:
    gearup(main)(['f1', '1'])
  except (TypeError, ValueError):
    pass
  except:
    raise

  try:
    gearup(main)(['f2', '1'])
  except (TypeError, ValueError):
    pass
  except:
    raise

def test_member():
  class A:
    class B:
      @staticmethod
      def f(x: int):
        return x

      @staticmethod
      def g(x: int):
        return x + 1

    @staticmethod
    def h(x: int):
      return x + 2

  def main(method: member[A], x: int):
    return method(x)

  assert gearup(main)(['B.f', '1']) == 1
  assert gearup(main)(['B.g', '1']) == 2
  assert gearup(main)(['h', '1']) == 3

  try:
    gearup(main)(['f1', '1'])
  except (TypeError, ValueError):
    pass
  except:
    raise

  try:
    gearup(main)(['f2', '1'])
  except (TypeError, ValueError):
    pass
  except:
    raise

def test_apply():
  def g1(x, y, z: float):
    return x + y + z

  def g2(x, y, h: float, jj: str):
    return (x + y) / h + len(jj)

  def main(f: choice(g1, g2), x: int, y: float, **kwargs):
    return apply(f, x, y, **kwargs)

  result = gearup(main)(['g1', 'x=1', 'y=2.0', 'z=1e-1'])
  assert result == 3.1

  result = gearup(main)(['g2', 'x=1', 'y=2', 'jj=1e-1', 'h=10'])
  assert result == 4.3

  try:
    gearup(main)(['g2', 'x=1', 'y=2', 'z=1e-1'])
  except (ValueError, TypeError) as e:
    pass
  except:
    raise

def test_kwargs():
  def f1(x):
    return x + 1

  def f2(x, flag: bool=False):
    return x if flag else -x

  def g1(x, alpha: float):
    return alpha * x

  def g2(x, beta: float):
    return x / beta

  def main(f: choice(f1, f2), g: choice(g1, g2), x: float, **kwargs):
    y = apply(g, x, **kwargs.get('gconfig', dict()))
    return apply(f, y, **kwargs.get('fconfig', dict()))

  result = gearup(main)(['f=f1', 'g=g2', 'x=1', 'gconfig.beta=2.0'])
  assert result == 1.5

  result = gearup(main)(['f=f1', 'g=g1', 'x=1', 'gconfig.alpha=2.0'])
  assert result == 3.0

  result = gearup(main)(['f=f2', 'g=g1', 'x=1', 'gconfig.alpha=2.0'])
  assert result == -2.0

  result = gearup(main)(['f=f2', 'g=g1', 'x=1', 'fconfig.flag=True', 'gconfig.alpha=3.0'])
  assert result == 3.0

  try:
    gearup(main)(['f=f1', 'g=g1', 'x=1', 'gconfig.flag=True'])
  except (ValueError, TypeError) as e:
    pass
  except:
    raise

def test_number():
  def check(interval, x, y, z):
    print(interval)

    if x is not None:
      try:
        interval(str(x))
      except (TypeError, ValueError) as e:
        pass
      except:
        raise

    if y is not None:
      assert interval(str(y)) == y

    if z is not None:
      try:
        interval(str(z))
      except (TypeError, ValueError) as e:
        pass
      except:
        raise


  print()
  check(1 < number, 1, 5, None)
  check(1 <= number, 0, 5, None)
  check(number < 10, None, 5, 10)
  check(number <= 10, None, 5, 11)
  check((1 <= number) & (number <= 10), 0, 5, 11)
  check((1 <= number) & (number < 10), 0, 5, 10)
  check((1 < number) & (number <= 10), 1, 5, 11)
  check((1 < number) & (number < 10), 1, 5, 10)

  print()
  check(interval[1:10], 0, 5, 10)
  check((1 < number) < 10, 1, 5, 10)


from typing import Union

import inspect
import os

import yaml
from omegaconf import OmegaConf

from .utils import indent_second

__all__ = [
  'gearup',
  'apply',
]

def apply(f):
  """
  Applies function `f` to the provided arguments and keyword arguments,
  ignoring keyword arguments that are not accepted by the function.

  Note: positional arguments `args` are always passed to the function.
  """
  def call(*args, **kwargs):
    signature = inspect.signature(f)

    passed_kwargs = dict()

    for name, p in signature.parameters.items():
      if p.kind == inspect.Parameter.POSITIONAL_ONLY:
        pass

      elif p.kind == inspect.Parameter.VAR_POSITIONAL:
        pass

      elif p.kind == inspect.Parameter.VAR_KEYWORD:
        ### function accepts everything
        return f(*args, **kwargs)

      elif name in kwargs:
        passed_kwargs[name] = kwargs[name]

    return f(*args, **passed_kwargs)

  return call

class cli_function(object):
  def __init__(self, f, helper: Union[None, str]=None, config: Union[None, bytes, str, os.PathLike]=None):
    self.f = f
    self.helper = helper
    self.config = config

  def help(self):
    doc: Union[None, str] = getattr(self.f, '__doc__', None)
    signature = str(inspect.signature(self.f))

    if doc is None:
      return f'{signature}'
    else:
      return f'{signature}\n{inspect.cleandoc(doc)}'

  def short_help(self):
    doc = getattr(self.f, '__doc__', None)
    signature = str(inspect.signature(self.f))

    if doc is None:
      return str(inspect.signature(self.f))
    else:
      header, *_ = inspect.cleandoc(doc).split('\n\n', maxsplit=1)

      return f'{signature}\n{header.strip()}'

  def __call__(self, arguments=None):
    if arguments is None:
      import sys
      arguments = sys.argv[1:]

    if len(arguments) == 1:
      if self.helper is not None and arguments[0] == self.helper:
        print(self.help())
        return

    if self.config is None:
      config = OmegaConf.create()
    else:
      config = OmegaConf.load(self.config)

    args = list()
    kwargs = list()

    for arg in arguments:
      if '=' in arg:
        kwargs.append(arg)
      else:
        value = yaml.load(arg, Loader=yaml.SafeLoader)
        args.append(value)

    cli_arguments = OmegaConf.from_dotlist(kwargs)
    ### CLI arguments have priority
    merged = OmegaConf.merge(config, cli_arguments)

    return apply(self.f)(*args, **merged)

class cli_commands(object):
  def __init__(self, commands, helper: Union[None, str], config: Union[None, bytes, str, os.PathLike]):
    self.commands = commands
    self.helper = helper
    self.config = config

  def signature(self):
    return '\n'.join(
      '%s -> %s' % (k, indent_second(v.short_help(), len(k) + 4))
      for k, v in self.commands.items()
    )

  def help(self):
    return (f'Available commands:\n'
            f'{self.signature()}')

  def short_help(self):
    return self.signature()

  def __call__(self, arguments=None):
    if arguments is None:
      import sys
      arguments = sys.argv[1:]

    if len(arguments) == 0:
      raise ValueError(
        f'please, specify command.\n'
        f'{self.help()}'
      )

    if len(arguments) == 1:
      if self.helper is not None and arguments[0] == self.helper:
        print(self.help())
        return

    command = arguments[0]
    if command not in self.commands:
      raise ValueError(
        f'invalid command {command}.\n'
        f'{self.help()}'
      )

    return self.commands[command](arguments[1:])

def compose(
  functions,
  named_functions,
  helper: Union[None, str]='help',
  config: Union[None, bytes, str, os.PathLike]=None
):
  if len(named_functions) == 0 and len(functions) == 0:
    raise ValueError('A function or dictionary of functions must be supplied!')

  if len(named_functions) == 0 and len(functions) == 1:
    return cli_function(functions[0], helper=helper, config=config)

  commands = dict()

  for function in functions:
    if hasattr(function, '__name__') and callable(function):
      name = function.__name__
      if name in commands:
        raise ValueError('duplicated command names (%s)' % (name,))
      else:
        commands[name] = cli_function(function, helper=helper, config=config)

    else:
      raise ValueError(
        f'non-keyword arguments must be callables and have `__name__` attribute (like all functions). '
        f'Got {function}'
      )

  for name, function in named_functions.items():
    if name in commands:
      raise ValueError('duplicated command names (%s)' % (name,))

    if isinstance(function, (list, tuple)):
      commands[name] = compose(function, dict(), helper=helper, config=config)
    elif isinstance(function, (dict,)):
      commands[name] = compose((), function, helper=helper, config=config)
    elif callable(function):
      commands[name] = cli_function(function, helper=helper, config=config)
    else:
      raise ValueError(
        f'command {name}={function} is not understood, must be either a callable, a tuple/list or a dict.'
      )

  return cli_commands(commands, helper=helper, config=config)

class GearUp(object):
  def __init__(
    self,
    functions, named_functions,
    helper: Union[None, str]='help',
    config: Union[None, bytes, str, os.PathLike]=None
  ):
    self.functions = functions
    self.named_functions = named_functions

    self.helper = helper
    self.config = config

  def with_config(self, config: Union[None, bytes, str, os.PathLike]) -> 'GearUp':
    return GearUp(self.functions, self.named_functions, helper=self.helper, config=config)

  def with_helper(self, helper: Union[None, str]) -> 'GearUp':
    return GearUp(self.functions, self.named_functions, helper=helper, config=self.config)

  def without_helper(self) -> 'GearUp':
    return GearUp(self.functions, self.named_functions, helper=None, config=self.config)

  def __call__(self, args):
    return compose(self.functions, self.named_functions, helper=self.helper, config=self.config)(args)

def gearup(*args, **kwargs) -> GearUp:
  return GearUp(args, kwargs)